package fc.accessible;

import java.lang.reflect.Constructor;
import java.lang.reflect.Field;
import java.lang.reflect.Method;
import java.util.Properties;

public class FCAccessibleObject {

    public static byte METHOD_TYPE_GET = 0;
    public static byte METHOD_TYPE_SET = 1;

    private static String PROP_DESC          = ".desc";
    private static String PROP_VALUE         = ".value";
    private static String PROP_MINIMUM       = ".minimum";
    private static String PROP_MAXIMUM       = ".maximum";
    private static String PROP_STEP          = ".step";
    private static String PROP_STEP_MULTIPLY = ".stepMultiply";

    private static Class[] STRING_PARAM = new Class[] { String.class };
    private static String[] METHOD_PREFIXES_GET = { "is", "get" };
    private static String[] METHOD_PREFIXES_SET = { "set" };

    private FCAccessible accessible;

    private String   id;
    private Method[] methods;
    
    private String desc = null;

    private Object value;
    private Object defaultValue;

    private Object  minimum;
    private Object  maximum;
    private Object  step;
    private boolean stepMultiply;

    private boolean modified = true;

    public FCAccessibleObject( FCAccessible pAccessible, String pId, Properties pProps ) throws Exception {

        accessible   = pAccessible;
        
        id           = pId;
        methods      = getMethods( pId );

        desc = pProps.getProperty( pId + PROP_DESC );

        defaultValue = stringToType( pProps.getProperty( pId + PROP_VALUE ) );
        value        = defaultValue;

        minimum = resolveRangeValue( pProps.getProperty( pId + PROP_MINIMUM ), "MIN_VALUE" );
        maximum = resolveRangeValue( pProps.getProperty( pId + PROP_MAXIMUM ), "MAX_VALUE" );
        
        step = stringToType( pProps.getProperty( pId + PROP_STEP ) );
        stepMultiply = Boolean.parseBoolean( pProps.getProperty( pId + PROP_STEP_MULTIPLY ) );
        
//        System.out.println( pName + " " + minimum + " " + maximum + " " + step );
        
        // -- Set default value.
//        clear();
        
        // -- Apply value on accessible object.
        apply();

    }

    private String buildExceptionMessage( String pName, String[] pPrefixes,
            Class pParam ) {

        StringBuffer wEm = new StringBuffer();

        for( int a = 0; a < pPrefixes.length; a++ ) {

            String wName = buildMethodName( pPrefixes[ a ], pName );

            if ( a > 0 ) {
                wEm.append( ", " );
            }

            wEm.append( wName + "(" );

            if ( pParam != null ) {
                wEm.append( " " + pParam.getName() + " " );
            }

            wEm.append( ")" );

        }

        return wEm.toString();

    }

    private String buildMethodName( String pPrefix, String pName ) {
        return pPrefix + Character.toUpperCase( pName.charAt( 0 ) ) +
            pName.substring( 1 );
    }

    public void clear() {
        setValue( defaultValue );
    }

    public void decrease() {
    	
        Class wType = getType();

        if ( wType.isPrimitive() ) {

            if ( wType == Byte.TYPE ) {
            	value = stepMultiply ? (byte)( (Byte)value / (Byte)step ) : (byte)( (Byte)value - (Byte)step );
            } else if ( wType == Short.TYPE ) {
                value = stepMultiply ? (short)( (Short)value / (Short)step ) : (short)( (Short)value - (Short)step );
            } else if ( wType == Integer.TYPE ) {
                value = stepMultiply ? (int)( (Integer)value / (Integer)step ) : (int)( (Integer)value - (Integer)step );
            } else if ( wType == Long.TYPE ) {
                value = stepMultiply ? (long)( (Long)value / (Long)step ) : (long)( (Long)value - (Long)step );
            } else if ( wType == Float.TYPE ) {
                value = stepMultiply ? (float)( (Float)value / (Float)step ) : (float)( (Float)value - (Float)step );
            } else if ( wType == Double.TYPE ) {
                value = stepMultiply ? (double)( (Double)value / (Double)step ) : (double)( (Double)value - (Double)step );
            } else if ( wType == Boolean.TYPE ) {
                value = !(Boolean)value;
            }

        }

        if ( ( (Comparable)value ).compareTo( minimum ) == -1 ) {
        	value = minimum;
        }
        
        modified = true;
        
//        apply();
        
    }
    
    private Method findMethod( String pName, String[] pPrefixes, Class pParam ) throws NoSuchMethodException {

		Class[] wParamArray = null;
		
		if ( pParam != null ) {
		    wParamArray = new Class[] { pParam };
		}
		
		for( int a = 0; a < pPrefixes.length; a++ ) {
		
		    String wName = buildMethodName( pPrefixes[ a ], pName );
		
		    try {
		        return accessible.getClass().getMethod( wName, wParamArray );
		    } catch( Exception pEx ) { 
		    	//pEx.printStackTrace();
		    }
		
		}
		
		throw new NoSuchMethodException(
		    buildExceptionMessage( pName, pPrefixes, pParam ) );
		
	}

    public Object getAccessible() {
        return accessible;
    }

    public Object getDefaultValue() {
    	return defaultValue;
    }
    
    public String getDescription() {
        return desc;
    }
    
    public String getId() {
    	return id;
    }

    protected Method getMethod( byte pType ) {
        return methods[ pType ];
    }

    private Method[] getMethods( String pName ) throws NoSuchMethodException {

        Method[] wMethods = new Method[ 2 ];

        // -- Find suitable getter method
        wMethods[ METHOD_TYPE_GET ] = findMethod(
            pName,
            METHOD_PREFIXES_GET,
            null );

        // -- Find matching setter method
        wMethods[ METHOD_TYPE_SET ] = findMethod(
            pName,
            METHOD_PREFIXES_SET,
            wMethods[ METHOD_TYPE_GET ].getReturnType() );

        return wMethods;

    }

    public Class getType() {
        return methods[ METHOD_TYPE_GET ].getReturnType();
    }
    
    public Object getMaximum() {
    	return maximum;
    }
    
    public Object getMinimum() {
    	return minimum;
    }
    
    public Object getStep() {
    	return step;
    }

    public Object getValue() {
//    	return getValue( true );
//    }
//    
//    public Object getValue( boolean pClearModified ) {
        
  // 		refresh();
        
        return value;
    }

    public void increase() {
    	
        Class wType = getType();

        if ( wType.isPrimitive() ) {

            if ( wType == Byte.TYPE ) {
                value = stepMultiply ? (byte)( (Byte)value * (Byte)step ) : (byte)( (Byte)value + (Byte)step );
            } else if ( wType == Short.TYPE ) {
                value = stepMultiply ? (short)( (Short)value * (Short)step ) : (short)( (Short)value + (Short)step );
            } else if ( wType == Integer.TYPE ) {
                value = stepMultiply ? (int)( (Integer)value * (Integer)step ) : (int)( (Integer)value + (Integer)step );
            } else if ( wType == Long.TYPE ) {
                value = stepMultiply ? (long)( (Long)value * (Long)step ) : (long)( (Long)value + (Long)step );
            } else if ( wType == Float.TYPE ) {
                value = stepMultiply ? (float)( (Float)value * (Float)step ) : (float)( (Float)value + (Float)step );
            } else if ( wType == Double.TYPE ) {
                value = stepMultiply ? (double)( (Double)value * (Double)step ) : (double)( (Double)value + (Double)step );
            } else if ( wType == Boolean.TYPE ) {
                value = !(Boolean)value;
            }

        }
        
        if ( ( (Comparable)value ).compareTo( maximum ) == 1 ) {
        	value = maximum;
        }
        
        modified = true;
        
//        apply();
    	
    }
    
    public boolean isModified() {
        return modified;
    }
    
    public boolean isStepMultiplied() {
    	return stepMultiply;
    }

    protected Object resolveRangeValue( String pValue, String pRangeType ) {
    	
    	if ( pValue != null ) {
    		return stringToType( pValue );
    	} else {
    		
            Class wType = getType();

            if ( wType.isPrimitive() ) {

				if ( wType == Byte.TYPE ) {
				    wType = Byte.class;
				} else if ( wType == Short.TYPE ) {
				    wType = Short.class;
				} else if ( wType == Integer.TYPE ) {
				    wType = Integer.class;
				} else if ( wType == Long.TYPE ) {
				    wType = Long.class;
				} else if ( wType == Float.TYPE ) {
				    wType = Float.class;
				} else if ( wType == Double.TYPE ) {
				    wType = Double.class;
				} else if ( wType == Boolean.TYPE ) {
					wType = Boolean.class;
				}

            }
            	
            if ( wType != Boolean.class ) {
        	
            	try {
            		
	            	Field wMvf = wType.getDeclaredField( pRangeType );
	            	
	            	return wMvf.get( null );
	            	
            	} catch( Exception pEx ) {
            		pEx.printStackTrace();
            	}
            	
            } else {
            	
            	// -- Handle booleans.
            	if ( pRangeType.equals( "MAX_VALUE" ) ) {
            		return Boolean.TRUE;
            	} else {
            		return Boolean.FALSE;
            	}
            	
            }

    	}
    	
        return null;
		
    }
    
    public void setMaximum( Object pMaximum ) {
    	maximum = pMaximum;
    }
    
    public void setMinimum( Object pMinimum ) {
    	minimum = pMinimum;
    }
    
    public void setModified( boolean pState ) {
    	
//    	if ( pState && !modified ) {
//    		System.out.println( getDescription() + " modified!" );
//    	} else if ( !pState && modified ) {
//    		System.out.println( getDescription() + " modified cleared!" );
//    	}
//    	
        modified = pState;
        
    }

    public void setStep( Object pStep ) {
    	step = pStep;
    }
    
    public void setValue( Object pValue ) {

        if ( !value.equals( pValue ) ) {
            
        	value = pValue;
            
            if ( ( (Comparable)value ).compareTo( minimum ) == -1 ) {
            	value = minimum;
            } else if ( ( (Comparable)value ).compareTo( maximum ) == 1 ) {
            	value = maximum;
            }
        	
            setModified( true );
            
        }

    }

    public void apply() {

    	if ( modified ) {
    		
	        try {
	            methods[ METHOD_TYPE_SET ].invoke( accessible, new Object[] { value } );
	        } catch( Exception pEx ) {
	            System.out.println( "apply: " + pEx );
	        }

	        modified = false;
	        
    	}

    }

    public void refresh() {

        try {
            value = methods[ METHOD_TYPE_GET ].invoke( accessible, null );
        } catch( Exception pEx ) {
            System.out.println( "refresh: " + pEx );
        }

    }

    public Object stringToType( String pStrValue ) {

//        System.out.println( "Desc: " + getDescription() );

        Class wType = getType();

        if ( wType.isPrimitive() ) {

        	if ( pStrValue == null || pStrValue.isEmpty() ) {
        		pStrValue = "0";
        	}
        	
            if ( wType == Byte.TYPE ) {
                return new Byte( pStrValue );
            } else if ( wType == Short.TYPE ) {
                return new Short( pStrValue );
            } else if ( wType == Integer.TYPE ) {
                return new Integer( pStrValue );
            } else if ( wType == Long.TYPE ) {
                return new Long( pStrValue );
            } else if ( wType == Float.TYPE ) {
                return new Float( pStrValue );
            } else if ( wType == Double.TYPE ) {
                return new Double( pStrValue );
            } else if ( wType == Boolean.TYPE ) {
                return new Boolean( pStrValue );
            }

        }

        try {
        
	        Constructor wCon = wType.getConstructor( STRING_PARAM );
	
	        return wCon.newInstance( new String[] { pStrValue } );
	        
        } catch( Exception pEx ) {
        	throw new FCAccessibleException( pEx );
        }

    }

}