/* ----------------------------------------------------------------------------
 * File:       fc/renderer/FCSolidGuessRenderer.java
 *
 * Author:     Unknown
 *
 * Copyright (c) - 1999, Technologies Interactives Mediagrif Inc.
 *
 * History
 * ----------------------------------------------------------------------------
 *  Date         User               Description
 * ----------------------------------------------------------------------------
 * 2000/02/18   Kris Fudalewski     Creation.
 * ----------------------------------------------------------------------------
 */
package fc.renderer;

import java.awt.Graphics;
import java.awt.Rectangle;
import java.util.Properties;

public class FCSolidGuessRenderer extends FCRenderer {

    public static final int SGM_TOP_LEFT     = 1;
    public static final int SGM_TOP_RIGHT    = 2;
    public static final int SGM_BOTTOM_LEFT  = 4;
    public static final int SGM_BOTTOM_RIGHT = 8;

    private short[][] map;
    private short[] filler = new short[ 256 ];

    protected int minGuessSize;
    protected int maxGuessSize;
    protected int passStep;
    
    protected boolean renderDetails;

    public FCSolidGuessRenderer( Properties pProp ) {
        super( pProp );
        setDetailLevel( 1 );
    }

    public boolean clearBeforeRender() {
        return true;
    }

    public short computeSolidGuess( Graphics pGrp, int pX, int pY, int pSz ) {

        int wHSz = pSz >> 1;

        // -- Compute five common points
        short wCn = mapPoint( pGrp, pX + wHSz, pY + wHSz ); // -- Center

        if ( wCn == mapPoint( pGrp, pX, pY ) ) {
            if ( wCn == mapPoint( pGrp, pX + pSz, pY ) ) {
                if ( wCn == mapPoint( pGrp, pX, pY + pSz ) ) {
                    if ( wCn == mapPoint( pGrp, pX + pSz, pY + pSz ) ) {
                        return wCn;
                    }
                }
            }
        }

        return -1;

    }

    private void createMap( int pWidth, int pHeight ) {
        map = new short[ pHeight ][ pWidth ];
    }

    public void drawSolidGuess( Graphics pGrp, int pX, int pY, int pSz ) {

        Rectangle wBounds = pGrp.getClipBounds();

        int wX = wBounds.x + pX;
        int wY = wBounds.y + pY;

        int wDth = pSz >> 1;
        short wItr;

        if ( ( wItr = computeSolidGuess( pGrp, pX, pY, pSz ) ) != -1 ) {
            fillMap( pX, pY, pSz, wItr );
            drawBlock( pGrp, wX, wY, pSz, wItr );
        }

    }

    private void fillMap( int pX, int pY, int pSz, short pItr ) {

        // -- Init filler
        if ( filler[ 0 ] != pItr ) {
            for ( int a = 0; a < pSz; a++ ) {
                filler[ a ] = pItr;
            }
        }

        for ( int a = pY; a < pY + pSz; a++ ) {
            System.arraycopy( filler, 0, map[ a ], pX, pSz );
        }

    }

    public int getMinimumGuessSize() {
        return minGuessSize;
    }

    public int getMaximumGuessSize() {
        return maxGuessSize;
    }

    public int getPassStep() {
        return passStep;
    }

    public boolean getRenderDetails() {
    	return renderDetails;
    }
    
    public short mapPoint( Graphics pGrp, int pX, int pY ) {

        Rectangle wBounds = pGrp.getClipBounds();

        int wX = wBounds.x + pX;
        int wY = wBounds.y + pY;

        if ( map[ pY ][ pX ] == 0 ) {

            short i = getFractal().computeViewPoint( wX, wY, getViewportSize().width, getViewportSize().height );

            drawPoint( pGrp, wX, wY, i );

            map[ pY ][ pX ] = i;

            return i;

        } else {
            return map[ pY ][ pX ];
        }

    }

    public void paint( Graphics pGrp ) {

        super.paint( pGrp );

        stop = false;
        
        // -- Get clipping bounds
        Rectangle wBounds = pGrp.getClipBounds();

        createMap( wBounds.width, wBounds.height );

        // -- Solid guessing loop
        for ( int d = maxGuessSize; d >= minGuessSize; d -= passStep ) {

            for ( int y = 0; y < wBounds.height - d; y += d ) {

                for( int x = 0; x < wBounds.width - d; x += d ) {

                    drawSolidGuess( pGrp, x, y, d );

                }

                if ( stop ) {
                	break;
                } else {
                	Thread.yield();
                }
                
            }

        }

        // -- Render details
        if ( renderDetails ) {
        	
	        for ( int y = 0; y < wBounds.height; y++ ) {
	
	            for( int x = 0; x < wBounds.width; x++ ) {
	
	                if ( map[ y ][ x ] == 0 ) {
	                    mapPoint( pGrp, x, y );
	                }
	
	            }
	
	            if ( stop ) {
	            	break;
	            } else {
	            	Thread.yield();
	            }
	            
	        }
	        
        }

    }

    public void setRenderDetails( boolean pState ) {
    	renderDetails = pState;
    }
    
    public void setMinimumGuessSize( int pSize ) {
        minGuessSize = pSize;
    }

    public void setMaximumGuessSize( int pSize ) {
        maxGuessSize = pSize;
    }

    public void setPassStep( int pStep ) {
        passStep = pStep;
    }

//    private class SolidGuess {
//
//        public int mask;
//        public short iterations;
//
//        public SolidGuess( int pMask, short pIterations ) {
//
//            mask = pMask;
//            iterations = pIterations;
//
//        }
//
//    }

}
